/***************************************************************************
 *   Copyright (C) 2008 by Ari Loytynoja   *
 *   ari@ebi.ac.uk   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef TRANSLATESEQUENCES_H
#define TRANSLATESEQUENCES_H

#include <string>
#include <vector>
#include <map>

/**
	@author Ari Loytynoja <ari@ebi.ac.uk>
*/

class TranslateSequences
{
//    std::map<std::string,std::string> dnaSeqs;
    std::map<std::string,std::string> codonToAa;
    std::map<std::string,std::string> aaToCodon;
public:
    TranslateSequences();

    ~TranslateSequences();
    bool translateProtein(const std::vector<std::string> *names,std::vector<std::string> *sequences,std::map<std::string,std::string> *dnaSequences);
    bool translateDNA(std::vector<std::string> *names,std::vector<std::string> *protein,std::vector<std::string> *dna,std::map<std::string,std::string> *dnaSequences);

//    std::map<std::string,std::string> getDnaSeqs() {return dnaSeqs; }
//    void setDnaSeqs(std::map<std::string,std::string> ds) {dnaSeqs = ds; }
};

#endif
